import subprocess

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from sublime.server import Server
from sublime.config import ServerConfiguration
from sublime.ui.common import EditFormDialog, IconButton


class EditServerDialog(EditFormDialog):
    entity_name: str = 'Server'
    initial_size = (450, 250)
    text_fields = [
        ('Name', 'name', False),
        ('Server address', 'server_address', False),
        ('Local network address', 'local_network_address', False),
        ('Local network SSID', 'local_network_ssid', False),
        ('Username', 'username', False),
        ('Password', 'password', True),
    ]
    boolean_fields = [
        ('Browse by tags', 'browse_by_tags'),
        ('Sync enabled', 'sync_enabled'),
        ('Do not verify certificate', 'disable_cert_verify'),
    ]

    def __init__(self, *args, **kwargs):
        test_server = Gtk.Button(label='Test Connection to Server')
        test_server.connect('clicked', self.on_test_server_clicked)

        open_in_browser = Gtk.Button(label='Open in Browser')
        open_in_browser.connect('clicked', self.on_open_in_browser_clicked)

        self.extra_buttons = [(test_server, None), (open_in_browser, None)]

        super().__init__(*args, **kwargs)

    def on_test_server_clicked(self, event):
        # Instantiate the server.
        server_address = self.data['server_address'].get_text()
        server = Server(
            name=self.data['name'].get_text(),
            hostname=server_address,
            username=self.data['username'].get_text(),
            password=self.data['password'].get_text(),
            disable_cert_verify=self.data['disable_cert_verify'].get_active(),
        )

        # Try to ping, and show a message box with whether or not it worked.
        try:
            server.ping()
            dialog = Gtk.MessageDialog(
                transient_for=self,
                message_type=Gtk.MessageType.INFO,
                buttons=Gtk.ButtonsType.OK,
                text='Connection to server successful.',
            )
            dialog.format_secondary_markup(
                f'Connection to {server_address} successful.')
        except Exception as err:
            dialog = Gtk.MessageDialog(
                transient_for=self,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.OK,
                text='Connection to server unsuccessful.',
            )
            dialog.format_secondary_markup(
                f'Connection to {server_address} resulted in the following '
                f'error:\n\n{err}')

        dialog.run()
        dialog.destroy()

    def on_open_in_browser_clicked(self, event):
        subprocess.call(['xdg-open', self.data['server_address'].get_text()])


class ConfigureServersDialog(Gtk.Dialog):
    __gsignals__ = {
        'server-list-changed':
        (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (object, )),
        'connected-server-changed':
        (GObject.SignalFlags.RUN_FIRST, GObject.TYPE_NONE, (object, )),
    }

    def __init__(self, parent, config):
        Gtk.Dialog.__init__(
            self,
            title='Configure Servers',
            transient_for=parent,
            flags=0,
            add_buttons=(),
        )

        self.server_configs = config.servers
        self.selected_server_index = config.current_server
        self.set_default_size(500, 300)

        # Flow box to hold the server list and the buttons.
        flowbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Server List
        self.server_list = Gtk.ListBox(activate_on_single_click=False)
        self.server_list.connect(
            'selected-rows-changed', self.server_list_on_selected_rows_changed)
        self.server_list.connect('row-activated', self.on_server_list_activate)
        flowbox.pack_start(self.server_list, True, True, 10)

        button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        # Tuples: (button, action function, pack side, requires_selection)
        # Add all of the buttons to the button box.
        self.buttons = [
            (
                IconButton(
                    'document-edit-symbolic',
                    label='Edit...',
                    relief=True,
                ), lambda e: self.on_edit_clicked(e, False), 'start', True),
            (
                IconButton(
                    'list-add-symbolic',
                    label='Add...',
                    relief=True,
                ), lambda e: self.on_edit_clicked(e, True), 'start', False),
            (
                IconButton(
                    'list-remove-symbolic',
                    label='Remove',
                    relief=True,
                ), self.on_remove_clicked, 'start', True),
            (
                IconButton(
                    'window-close-symbolic',
                    label='Close',
                    relief=True,
                ), lambda _: self.close(), 'end', False),
            (
                IconButton(
                    'network-transmit-receive-symbolic',
                    label='Connect',
                    relief=True,
                ), self.on_connect_clicked, 'end', True),
        ]
        for button_cfg in self.buttons:
            btn, action, pack_end, requires_selection = button_cfg

            if pack_end == 'end':
                button_box.pack_end(btn, False, True, 5)
            else:
                button_box.pack_start(btn, False, True, 5)

            btn.connect('clicked', action)

        flowbox.pack_end(button_box, False, False, 0)

        # Add the flowbox to the dialog and show the dialog.
        content_area = self.get_content_area()
        content_area.pack_start(flowbox, True, True, 10)

        self.show_all()
        self.refresh_server_list()
        self.server_list_on_selected_rows_changed(None)

    def refresh_server_list(self):
        # Clear out the list.
        for el in self.server_list:
            self.server_list.remove(el)

        # Add all of the rows for each of the servers.
        for i, config in enumerate(self.server_configs):
            box = Gtk.Box()
            image = Gtk.Image(margin=5)
            if i == self.selected_server_index:
                image.set_from_icon_name(
                    'network-transmit-receive-symbolic',
                    Gtk.IconSize.SMALL_TOOLBAR,
                )

            box.add(image)

            server_name_label = Gtk.Label(label=config.name)
            server_name_label.set_halign(Gtk.Align.START)
            box.add(server_name_label)
            self.server_list.add(box)

        # Show them, and select the current server.
        self.show_all()
        if (self.selected_server_index is not None
                and self.selected_server_index >= 0):
            self.server_list.select_row(
                self.server_list.get_row_at_index(self.selected_server_index))

    def on_remove_clicked(self, event):
        selected = self.server_list.get_selected_row()
        if selected:
            del self.server_configs[selected.get_index()]
            self.refresh_server_list()
            self.emit('server-list-changed', self.server_configs)

    def on_edit_clicked(self, event, add):
        if add:
            dialog = EditServerDialog(self)
        else:
            selected_index = self.server_list.get_selected_row().get_index()
            dialog = EditServerDialog(
                self, self.server_configs[selected_index])

        result = dialog.run()
        if result == Gtk.ResponseType.OK:
            # Create a new server configuration to use.
            new_config = ServerConfiguration(
                name=dialog.data['name'].get_text(),
                server_address=dialog.data['server_address'].get_text(),
                local_network_address=dialog.data['local_network_address']
                .get_text(),
                local_network_ssid=dialog.data['local_network_ssid'].get_text(
                ),
                username=dialog.data['username'].get_text(),
                password=dialog.data['password'].get_text(),
                browse_by_tags=dialog.data['browse_by_tags'].get_active(),
                sync_enabled=dialog.data['sync_enabled'].get_active(),
                disable_cert_verify=dialog.data['disable_cert_verify']
                .get_active(),
            )

            if add:
                self.server_configs.append(new_config)
            else:
                self.server_configs[selected_index] = new_config

            self.refresh_server_list()
            self.emit('server-list-changed', self.server_configs)

        dialog.destroy()

    def on_server_list_activate(self, *args):
        self.on_connect_clicked(None)

    def on_connect_clicked(self, event):
        selected_index = self.server_list.get_selected_row().get_index()
        self.emit('connected-server-changed', selected_index)
        self.close()

    def server_list_on_selected_rows_changed(self, event):
        # Update the state of the buttons depending on whether or not a row is
        # selected in the server list.
        has_selection = self.server_list.get_selected_row()

        for button, *_, requires_selection in self.buttons:
            button.set_sensitive(not requires_selection or has_selection)



import os
from enum import Enum
import json
from typing import List

from .from_json import from_json
from .config import AppConfiguration
from .cache_manager import CacheManager
from .server.api_objects import Child


class RepeatType(Enum):
    NO_REPEAT = 0
    REPEAT_QUEUE = 1
    REPEAT_SONG = 2

    @property
    def icon(self):
        icon_name = [
            'repeat',
            'repeat-symbolic',
            'repeat-song-symbolic',
        ][self.value]
        return 'media-playlist-' + icon_name

    def as_mpris_loop_status(self):
        return ['None', 'Playlist', 'Track'][self.value]

    @staticmethod
    def from_mpris_loop_status(loop_status):
        return {
            'None': RepeatType.NO_REPEAT,
            'Track': RepeatType.REPEAT_SONG,
            'Playlist': RepeatType.REPEAT_QUEUE,
        }[loop_status]


class ApplicationState:
    """
    Represents the state of the application. In general, there are two things
    that are stored here: configuration, and UI state.

    Configuration is stored in ``config`` which is an ``AppConfiguration``
    object. UI state is stored as separate properties on this class.

    Configuration is stored to disk in $XDG_CONFIG_HOME/sublime-music. State is
    stored in $XDG_CACHE_HOME. Nothing in state should be assumed to be
    permanent. State need not be saved, the ``to_json`` and ``from_json``
    functions define what part of the state will be saved across application
    loads.
    """
    config: AppConfiguration = AppConfiguration()
    current_song: Child = None
    config_file: str = None
    playing: bool = False
    play_queue: List[str] = []
    old_play_queue: List[str] = []
    _volume: dict = {'this device': 100}
    is_muted: bool = False
    repeat_type: RepeatType = RepeatType.NO_REPEAT
    shuffle_on: bool = False
    song_progress: float = 0
    current_device: str = 'this device'
    current_tab: str = 'albums'
    selected_album_id: str = None
    selected_artist_id: str = None
    selected_playlist_id: str = None

    # State for Album sort.
    current_album_sort: str = 'random'
    current_album_genre: str = 'Rock'
    current_album_alphabetical_sort: str = 'name'
    current_album_from_year: int = 2010
    current_album_to_year: int = 2020

    active_playlist_id: str = None

    def to_json(self):
        current_song = (
            self.current_song.id if
            (hasattr(self, 'current_song')
             and self.current_song is not None) else None)
        # TODO this really sucks. We should fix this.
        return {
            'current_song':
            current_song,
            'play_queue':
            getattr(self, 'play_queue', None),
            'old_play_queue':
            getattr(self, 'old_play_queue', None),
            '_volume':
            getattr(self, '_volume', {}),
            'is_muted':
            getattr(self, 'is_muted', None),
            'repeat_type':
            getattr(self, 'repeat_type', RepeatType.NO_REPEAT).value,
            'shuffle_on':
            getattr(self, 'shuffle_on', None),
            'song_progress':
            getattr(self, 'song_progress', None),
            'current_device':
            getattr(self, 'current_device', 'this device'),
            'current_tab':
            getattr(self, 'current_tab', 'albums'),
            'selected_album_id':
            getattr(self, 'selected_album_id', None),
            'selected_artist_id':
            getattr(self, 'selected_artist_id', None),
            'selected_playlist_id':
            getattr(self, 'selected_playlist_id', None),
            'current_album_sort':
            getattr(self, 'current_album_sort', None),
            'current_album_genre':
            getattr(self, 'current_album_genre', None),
            'current_album_alphabetical_sort':
            getattr(self, 'current_album_alphabetical_sort', None),
            'current_album_from_year':
            getattr(self, 'current_album_from_year', None),
            'current_album_to_year':
            getattr(self, 'current_album_to_year', None),
            'active_playlist_id':
            getattr(self, 'active_playlist_id', None),
        }

    def load_from_json(self, json_object):
        current_song_id = json_object.get('current_song') or None
        if current_song_id and CacheManager.cache:
            self.current_song = CacheManager.cache['song_details'].get(
                current_song_id)
        else:
            self.current_song = None

        self.play_queue = json_object.get('play_queue', [])
        self.old_play_queue = json_object.get('old_play_queue', [])
        self._volume = json_object.get('_volume', {'this device': 100})
        self.is_muted = json_object.get('is_muted', False)
        self.repeat_type = RepeatType(json_object.get('repeat_type', 0))
        self.shuffle_on = json_object.get('shuffle_on', False)
        self.song_progress = json_object.get('song_progress', 0.0)
        self.current_device = json_object.get('current_device', 'this device')
        self.current_tab = json_object.get('current_tab', 'albums')
        self.selected_album_id = json_object.get('selected_album_id', None)
        self.selected_artist_id = json_object.get('selected_artist_id', None)
        self.selected_playlist_id = json_object.get(
            'selected_playlist_id', None)
        self.current_album_sort = json_object.get(
            'current_album_sort', 'random')
        self.current_album_genre = json_object.get(
            'current_album_genre', 'Rock')
        self.current_album_alphabetical_sort = json_object.get(
            'current_album_alphabetical_sort', 'name')
        self.current_album_from_year = json_object.get(
            'current_album_from_year', 2010)
        self.current_album_to_year = json_object.get(
            'current_album_to_year', 2020)
        self.active_playlist_id = json_object.get('active_playlist_id', None)

    def load(self):
        self.config = self.get_config(self.config_file)

        if self.config.server is not None:
            # Reset the CacheManager.
            CacheManager.reset(self.config, self.config.server)

        if (self.config.server is not None
                and os.path.exists(self.state_filename)):
            with open(self.state_filename, 'r') as f:
                try:
                    self.load_from_json(json.load(f))
                except json.decoder.JSONDecodeError:
                    # Who cares, it's just state.
                    pass
        else:
            self.load_from_json({})

    def save(self):
        # Make the necessary directories before writing the state.
        os.makedirs(os.path.dirname(self.state_filename), exist_ok=True)

        # Save the state
        with open(self.state_filename, 'w+') as f:
            f.write(json.dumps(self.to_json(), indent=2, sort_keys=True))

    def save_config(self):
        # Make the necessary directories before writing the config.
        os.makedirs(os.path.dirname(self.config_file), exist_ok=True)

        with open(self.config_file, 'w+') as f:
            f.write(
                json.dumps(self.config.to_json(), indent=2, sort_keys=True))

    def get_config(self, filename: str) -> AppConfiguration:
        if not os.path.exists(filename):
            return AppConfiguration()

        with open(filename, 'r') as f:
            try:
                return from_json(AppConfiguration, json.load(f))
            except json.decoder.JSONDecodeError:
                return AppConfiguration()

    @property
    def state_filename(self):
        default_cache_location = (
            os.environ.get('XDG_DATA_HOME')
            or os.path.expanduser('~/.local/share'))
        return os.path.join(
            default_cache_location,
            'sublime-music',
            CacheManager.calculate_server_hash(self.config.server),
            'state.yaml',
        )

    @property
    def volume(self):
        return self._volume.get(self.current_device, 100)

    @volume.setter
    def volume(self, value):
        self._volume[self.current_device] = value



from .album_with_songs import AlbumWithSongs
from .edit_form_dialog import EditFormDialog
from .icon_button import IconButton
from .spinner_image import SpinnerImage

__all__ = (
    'AlbumWithSongs',
    'EditFormDialog',
    'IconButton',
    'SpinnerImage',
)



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk


class SpinnerImage(Gtk.Overlay):
    def __init__(
            self,
            loading=True,
            image_name=None,
            spinner_name=None,
            **kwargs,
    ):
        Gtk.Overlay.__init__(self)

        self.image = Gtk.Image(name=image_name, **kwargs)
        self.add(self.image)

        self.spinner = Gtk.Spinner(
            name=spinner_name,
            active=loading,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

    def set_from_file(self, *args, **kwargs):
        self.image.set_from_file(*args, **kwargs)

    def set_loading(self, loading_status):
        if loading_status:
            self.spinner.start()
            self.spinner.show()
        else:
            self.spinner.stop()
            self.spinner.hide()



"""
Defines the base class for API objects.
"""

from enum import Enum
from typing import Any, Dict

from sublime.from_json import from_json as _from_json


class APIObject:
    @classmethod
    def from_json(cls, data):
        return _from_json(cls, data)

    def get(self, field, default=None):
        return getattr(self, field, default)

    def __repr__(self):
        if isinstance(self, Enum):
            return super().__repr__()
        if isinstance(self, str):
            return self

        annotations: Dict[str, Any] = self.get('__annotations__', {})
        typename = type(self).__name__
        fieldstr = ' '.join(
            [
                f'{field}={getattr(self, field)!r}'
                for field in annotations.keys()
                if hasattr(self, field) and getattr(self, field) is not None
            ])
        return f'<{typename} {fieldstr}>'



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio, Gtk, GObject

from . import albums, artists, playlists, player_controls
from sublime.state_manager import ApplicationState


class MainWindow(Gtk.ApplicationWindow):
    """Defines the main window for Sublime Music."""
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (str, object, object),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_default_size(1150, 768)

        # Create the stack
        self.stack = self.create_stack(
            Albums=albums.AlbumsPanel(),
            Artists=artists.ArtistsPanel(),
            Playlists=playlists.PlaylistsPanel(),
        )
        self.stack.set_transition_type(
            Gtk.StackTransitionType.SLIDE_LEFT_RIGHT)

        self.titlebar = self.create_headerbar(self.stack)
        self.set_titlebar(self.titlebar)

        self.player_controls = player_controls.PlayerControls()
        self.player_controls.connect(
            'song-clicked', lambda _, *a: self.emit('song-clicked', *a))

        flowbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        flowbox.pack_start(self.stack, True, True, 0)
        flowbox.pack_start(self.player_controls, False, True, 0)
        self.add(flowbox)

    def update(self, state: ApplicationState, force=False):
        # Update the Connected to label on the popup menu.
        if state.config.current_server >= 0:
            server_name = state.config.servers[
                state.config.current_server].name
            self.connected_to_label.set_markup(
                f'<b>Connected to {server_name}</b>')
        else:
            self.connected_to_label.set_markup(
                f'<span style="italic">Not Connected to a Server</span>')

        self.stack.set_visible_child_name(state.current_tab)

        active_panel = self.stack.get_visible_child()
        if hasattr(active_panel, 'update'):
            active_panel.update(state, force=force)

        self.player_controls.update(state)

    def create_stack(self, **kwargs):
        stack = Gtk.Stack()
        for name, child in kwargs.items():
            child.connect(
                'song-clicked',
                lambda _, *args: self.emit('song-clicked', *args),
            )
            child.connect(
                'refresh-window',
                lambda _, *args: self.emit('refresh-window', *args),
            )
            stack.add_titled(child, name.lower(), name)
        return stack

    def create_headerbar(self, stack):
        """
        Configure the header bar for the window.
        """
        header = Gtk.HeaderBar()
        header.set_show_close_button(True)
        header.props.title = 'Sublime Music'

        # Search
        search = Gtk.SearchEntry()
        header.pack_start(search)

        # Stack switcher
        switcher = Gtk.StackSwitcher(stack=stack)
        header.set_custom_title(switcher)

        # Menu button
        button = Gtk.MenuButton()
        button.set_use_popover(True)
        button.set_popover(self.create_menu())
        button.connect('clicked', self.on_menu_click)

        icon = Gio.ThemedIcon(name="open-menu-symbolic")
        image = Gtk.Image.new_from_gicon(icon, Gtk.IconSize.BUTTON)
        button.add(image)

        header.pack_end(button)

        return header

    def create_menu(self):
        self.menu = Gtk.PopoverMenu()

        self.connected_to_label = Gtk.Label(name='connected-to-label')
        self.connected_to_label.set_markup(
            f'<span style="italic">Not Connected to a Server</span>')

        menu_items = [
            (None, self.connected_to_label),
            (
                'app.configure-servers',
                Gtk.ModelButton(text='Configure Servers'),
            ),
            ('app.settings', Gtk.ModelButton(text='Settings')),
        ]

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        for name, item in menu_items:
            if name:
                item.set_action_name(name)
            item.get_style_context().add_class('menu-button')
            vbox.pack_start(item, False, True, 0)
        self.menu.add(vbox)

        return self.menu

    # ========== Event Listeners ==========
    def on_menu_click(self, button):
        self.menu.set_relative_to(button)
        self.menu.show_all()
        self.menu.popup()



import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject

from .common.edit_form_dialog import EditFormDialog


class SettingsDialog(EditFormDialog):
    title: str = 'Settings'
    initial_size = (450, 250)
    text_fields = [
        ('Port Number (will take effect on restart)', 'port_number', False),
    ]
    boolean_fields = [
        ('Show headers on song lists', 'show_headers'),
        ('Always stream songs', 'always_stream'),
        ('When streaming, also download song', 'download_on_stream'),
        (
            'Show a notification when a song begins to play',
            'song_play_notification',
        ),
    ]
    numeric_fields = [
        (
            'How many songs in the play queue do you want to prefetch?',
            'prefetch_amount',
            (0, 10, 1),
            0,
        ),
        (
            'How many song downloads do you want to allow concurrently?',
            'concurrent_download_limit',
            (1, 10, 1),
            5,
        ),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import gi
from typing import Optional, Union

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GObject, GLib, Gio, Pango

from sublime.state_manager import ApplicationState
from sublime.cache_manager import CacheManager
from sublime.ui import util
from sublime.ui.common import AlbumWithSongs, IconButton, SpinnerImage

from sublime.server.api_objects import Child, AlbumWithSongsID3

Album = Union[Child, AlbumWithSongsID3]


class AlbumsPanel(Gtk.Box):
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (str, object, object),
        ),
        'refresh-window': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, bool),
        ),
    }

    populating_genre_combo = False

    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL)

        actionbar = Gtk.ActionBar()

        # Sort by
        actionbar.add(Gtk.Label(label='Sort'))
        self.sort_type_combo = self.make_combobox(
            (
                ('random', 'randomly'),
                ('byGenre', 'by genre'),
                ('newest', 'by most recently added'),
                ('highest', 'by highest rated'),
                ('frequent', 'by most played'),
                ('recent', 'by most recently played'),
                ('alphabetical', 'alphabetically'),
                ('starred', 'by starred only'),
                ('byYear', 'by year'),
            ),
            self.on_type_combo_changed,
        )
        actionbar.pack_start(self.sort_type_combo)

        # Alphabetically how?
        self.alphabetical_type_combo = self.make_combobox(
            (
                ('name', 'by album name'),
                ('artist', 'by artist name'),
            ),
            self.on_alphabetical_type_change,
        )
        actionbar.pack_start(self.alphabetical_type_combo)

        # Alphabetically how?
        self.genre_combo = self.make_combobox((), self.on_genre_change)
        actionbar.pack_start(self.genre_combo)

        self.from_year_label = Gtk.Label(label='from')
        actionbar.pack_start(self.from_year_label)
        self.from_year_entry = Gtk.Entry(input_purpose=Gtk.InputPurpose.DIGITS)
        self.from_year_entry.connect('changed', self.on_year_changed)
        actionbar.pack_start(self.from_year_entry)

        self.to_year_label = Gtk.Label(label='to')
        actionbar.pack_start(self.to_year_label)
        self.to_year_entry = Gtk.Entry(input_purpose=Gtk.InputPurpose.DIGITS)
        self.to_year_entry.connect('changed', self.on_year_changed)
        actionbar.pack_start(self.to_year_entry)

        refresh = IconButton('view-refresh-symbolic')
        refresh.connect('clicked', self.on_refresh_clicked)
        actionbar.pack_end(refresh)

        self.add(actionbar)

        scrolled_window = Gtk.ScrolledWindow()
        self.grid = AlbumsGrid()
        self.grid.connect(
            'song-clicked',
            lambda _, *args: self.emit('song-clicked', *args),
        )
        self.grid.connect('cover-clicked', self.on_grid_cover_clicked)
        scrolled_window.add(self.grid)
        self.add(scrolled_window)

    def make_combobox(self, items, on_change):
        store = Gtk.ListStore(str, str)
        for item in items:
            store.append(item)

        combo = Gtk.ComboBox.new_with_model(store)
        combo.set_id_column(0)
        combo.connect('changed', on_change)

        renderer_text = Gtk.CellRendererText()
        combo.pack_start(renderer_text, True)
        combo.add_attribute(renderer_text, 'text', 1)

        return combo

    def populate_genre_combo(
            self,
            state: ApplicationState,
            force: bool = False,
    ):
        if not CacheManager.ready():
            return

        def get_genres_done(f):
            try:
                new_store = [
                    (genre.value, genre.value) for genre in (f.result() or [])
                ]

                util.diff_song_store(self.genre_combo.get_model(), new_store)

                if self.get_id(self.genre_combo) != state.current_album_genre:
                    self.genre_combo.set_active_id(state.current_album_genre)
            finally:
                self.updating_query = False

        # Never force. We invalidate the cache ourselves (force is used when
        # sort params change).
        genres_future = CacheManager.get_genres(force=False)
        genres_future.add_done_callback(
            lambda f: GLib.idle_add(get_genres_done, f))

    def update(self, state: ApplicationState, force: bool = False):
        self.updating_query = True

        self.sort_type_combo.set_active_id(state.current_album_sort)
        self.alphabetical_type_combo.set_active_id(
            state.current_album_alphabetical_sort)
        self.from_year_entry.set_text(str(state.current_album_from_year))
        self.to_year_entry.set_text(str(state.current_album_to_year))
        self.populate_genre_combo(state, force=force)

        # Show/hide the combo boxes.
        def show_if(sort_type, *elements):
            for element in elements:
                if state.current_album_sort == sort_type:
                    element.show()
                else:
                    element.hide()

        show_if('alphabetical', self.alphabetical_type_combo)
        show_if('byGenre', self.genre_combo)
        show_if('byYear', self.from_year_label, self.from_year_entry)
        show_if('byYear', self.to_year_label, self.to_year_entry)

        self.grid.update(
            state=state,
            force=force,
            selected_id=state.selected_album_id,
        )

    def get_id(self, combo):
        tree_iter = combo.get_active_iter()
        if tree_iter is not None:
            return combo.get_model()[tree_iter][0]

    def on_refresh_clicked(self, button):
        self.emit('refresh-window', {}, True)

    def on_type_combo_changed(self, combo):
        new_active_sort = self.get_id(combo)
        self.grid.update_params(type_=new_active_sort)
        self.emit_if_not_updating(
            'refresh-window',
            {
                'current_album_sort': new_active_sort,
                'selected_album_id': None,
            },
            False,
        )

    def on_alphabetical_type_change(self, combo):
        new_active_alphabetical_sort = self.get_id(combo)
        self.grid.update_params(alphabetical_type=new_active_alphabetical_sort)
        self.emit_if_not_updating(
            'refresh-window',
            {
                'current_album_alphabetical_sort':
                new_active_alphabetical_sort,
                'selected_album_id': None,
            },
            False,
        )

    def on_genre_change(self, combo):
        new_active_genre = self.get_id(combo)
        self.grid.update_params(genre=new_active_genre)
        self.emit_if_not_updating(
            'refresh-window',
            {
                'current_album_genre': new_active_genre,
                'selected_album_id': None,
            },
            True,
        )

    def on_year_changed(self, entry):
        try:
            year = int(entry.get_text())
        except:
            # TODO prevent input of non-numerals
            print('failed, should do something to prevent non-numeric input')
            return

        if self.to_year_entry == entry:
            self.grid.update_params(to_year=year)
            self.emit_if_not_updating(
                'refresh-window',
                {
                    'current_album_to_year': year,
                    'selected_album_id': None,
                },
                True,
            )
        else:
            self.grid.update_params(from_year=year)
            self.emit_if_not_updating(
                'refresh-window',
                {
                    'current_album_from_year': year,
                    'selected_album_id': None,
                },
                True,
            )

    def on_grid_cover_clicked(self, grid, id):
        self.emit(
            'refresh-window',
            {'selected_album_id': id},
            False,
        )

    def emit_if_not_updating(self, *args):
        if self.updating_query:
            return
        self.emit(*args)


class AlbumModel(GObject.Object):
    def __init__(self, album: Album):
        self.album: Album = album
        super().__init__()

    @property
    def id(self):
        return self.album.id

    def __repr__(self):
        return f'<AlbumModel {self.album}>'


class AlbumsGrid(Gtk.Overlay):
    """Defines the albums panel."""
    __gsignals__ = {
        'song-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (str, object, object),
        ),
        'cover-clicked': (
            GObject.SignalFlags.RUN_FIRST,
            GObject.TYPE_NONE,
            (object, ),
        ),
    }
    type_: str = None
    alphabetical_type: str = 'name'
    from_year: int = 2010
    to_year: int = 2020
    genre: str = 'Rock'

    current_selection = None
    next_page_fn = None
    parameters_changed = False
    current_min_size_request = 30
    overshoot_update_in_progress = False
    server_hash = None

    def update_params(
            self,
            type_: str = None,
            alphabetical_type: str = None,
            from_year: int = None,
            to_year: int = None,
            genre: str = None,
    ):
        self.type_ = type_ or self.type_
        self.alphabetical_type = alphabetical_type or self.alphabetical_type
        self.from_year = from_year or self.from_year
        self.to_year = to_year or self.to_year
        self.genre = genre or self.genre
        self.parameters_changed = True
        self.current_min_size_request = 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # This is the master list.
        self.list_store = Gio.ListStore()

        self.items_per_row = 4

        scrolled_window = Gtk.ScrolledWindow()
        grid_detail_grid_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        self.grid_top = Gtk.FlowBox(
            hexpand=True,
            row_spacing=5,
            column_spacing=5,
            margin_top=12,
            margin_bottom=12,
            homogeneous=True,
            valign=Gtk.Align.START,
            halign=Gtk.Align.CENTER,
            selection_mode=Gtk.SelectionMode.SINGLE,
        )
        self.grid_top.connect('child-activated', self.on_child_activated)
        self.grid_top.connect('size-allocate', self.on_grid_resize)

        self.list_store_top = Gio.ListStore()
        self.grid_top.bind_model(self.list_store_top, self.create_widget)

        grid_detail_grid_box.add(self.grid_top)

        self.detail_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.detail_box.pack_start(Gtk.Box(), True, True, 0)

        self.detail_box_inner = Gtk.Box()
        self.detail_box.pack_start(self.detail_box_inner, False, False, 0)

        self.detail_box.pack_start(Gtk.Box(), True, True, 0)
        grid_detail_grid_box.add(self.detail_box)

        self.grid_bottom = Gtk.FlowBox(
            vexpand=True,
            hexpand=True,
            row_spacing=5,
            column_spacing=5,
            margin_top=12,
            margin_bottom=12,
            homogeneous=True,
            valign=Gtk.Align.START,
            halign=Gtk.Align.CENTER,
            selection_mode=Gtk.SelectionMode.SINGLE,
        )
        self.grid_bottom.connect('child-activated', self.on_child_activated)

        self.list_store_bottom = Gio.ListStore()
        self.grid_bottom.bind_model(self.list_store_bottom, self.create_widget)

        grid_detail_grid_box.add(self.grid_bottom)

        scrolled_window.add(grid_detail_grid_box)
        self.add(scrolled_window)

        self.spinner = Gtk.Spinner(
            name='grid-spinner',
            active=True,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

    def update(
            self,
            state: ApplicationState = None,
            force: bool = False,
            selected_id: str = None,
    ):
        new_hash = CacheManager.calculate_server_hash(state.config.server)
        if self.server_hash != new_hash:
            self.parameters_changed = True
        self.server_hash = new_hash

        self.update_grid(force=force, selected_id=selected_id)

        # Update the detail panel.
        children = self.detail_box_inner.get_children()
        if len(children) > 0 and hasattr(children[0], 'update'):
            children[0].update(force=force)

    error_dialog = None

    def update_grid(self, force=False, selected_id=None):
        if not CacheManager.ready():
            self.spinner.hide()
            return

        def reflow_grid(force_reload, selected_index):
            selection_changed = (selected_index != self.current_selection)
            self.current_selection = selected_index
            self.reflow_grids(
                force_reload_from_master=force_reload,
                selection_changed=selection_changed,
            )
            self.spinner.hide()
            self.overshoot_update_in_progress = False

        # Calculate the type.
        type_ = self.type_
        if self.type_ == 'alphabetical':
            type_ += {
                'name': 'ByName',
                'artist': 'ByArtist',
            }[self.alphabetical_type]

        require_reflow = self.parameters_changed
        self.parameters_changed = False

        def do_update(f):
            try:
                albums = f.result()
            except Exception as e:
                if self.error_dialog:
                    self.spinner.hide()
                    return
                self.error_dialog = Gtk.MessageDialog(
                    transient_for=self.get_toplevel(),
                    message_type=Gtk.MessageType.ERROR,
                    buttons=Gtk.ButtonsType.OK,
                    text='Failed to retrieve albums',
                )
                self.error_dialog.format_secondary_markup(
                    f'Getting albums by {type_} failed due to the following'
                    f' error\n\n{e}')
                self.error_dialog.run()
                self.error_dialog.destroy()
                self.error_dialog = None
                self.spinner.hide()
                return

            old_len = len(self.list_store)
            self.list_store.remove_all()

            selected_index = None
            for i, album in enumerate(albums):
                model = AlbumModel(album)

                if model.id == selected_id:
                    selected_index = i

                self.list_store.append(model)

            reflow_grid(
                force or require_reflow or (old_len != len(self.list_store)),
                selected_index,
            )

        future = CacheManager.get_album_list(
            type_=type_,
            from_year=self.from_year,
            to_year=self.to_year,
            genre=self.genre,
            before_download=lambda: GLib.idle_add(self.spinner.show),
            force=force,
        )
        future.add_done_callback(lambda f: GLib.idle_add(do_update, f))

    # Event Handlers
    # =========================================================================
    def on_child_activated(self, flowbox, child):
        click_top = flowbox == self.grid_top
        selected_index = (
            child.get_index() + (0 if click_top else len(self.list_store_top)))

        if selected_index == self.current_selection:
            self.emit('cover-clicked', None)
        else:
            self.emit('cover-clicked', self.list_store[selected_index].id)

    def on_grid_resize(self, flowbox, rect):
        # TODO: this doesn't work with themes that add extra padding.
        # 200     + (10      * 2) + (5      * 2) = 230
        # picture + (padding * 2) + (margin * 2)
        new_items_per_row = min((rect.width // 230), 7)
        if new_items_per_row != self.items_per_row:
            self.items_per_row = min((rect.width // 230), 7)
            self.detail_box_inner.set_size_request(
                self.items_per_row * 230 - 10,
                -1,
            )

            self.reflow_grids()

    # Helper Methods
    # =========================================================================
    def create_widget(self, item):
        widget_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        # Cover art image
        artwork = SpinnerImage(
            loading=False,
            image_name='grid-artwork',
            spinner_name='grid-artwork-spinner',
        )
        widget_box.pack_start(artwork, False, False, 0)

        def make_label(text, name):
            return Gtk.Label(
                name=name,
                label=text,
                tooltip_text=text,
                ellipsize=Pango.EllipsizeMode.END,
                max_width_chars=20,
                halign=Gtk.Align.START,
            )

        # Header for the widget
        header_text = (
            item.album.title if type(item.album) == Child else item.album.name)

        header_label = make_label(header_text, 'grid-header-label')
        widget_box.pack_start(header_label, False, False, 0)

        # Extra info for the widget
        info_text = util.dot_join(item.album.artist, item.album.year)
        if info_text:
            info_label = make_label(info_text, 'grid-info-label')
            widget_box.pack_start(info_label, False, False, 0)

        # Download the cover art.
        def on_artwork_downloaded(f):
            artwork.set_from_file(f.result())
            artwork.set_loading(False)

        def start_loading():
            artwork.set_loading(True)

        cover_art_filename_future = CacheManager.get_cover_art_filename(
            item.album.coverArt,
            before_download=lambda: GLib.idle_add(start_loading),
        )
        cover_art_filename_future.add_done_callback(
            lambda f: GLib.idle_add(on_artwork_downloaded, f))

        widget_box.show_all()
        return widget_box

    def reflow_grids(
            self,
            force_reload_from_master=False,
            selection_changed=False,
    ):
        # Determine where the cuttoff is between the top and bottom grids.
        entries_before_fold = len(self.list_store)
        if self.current_selection is not None and self.items_per_row:
            entries_before_fold = (
                ((self.current_selection // self.items_per_row) + 1)
                * self.items_per_row)

        if force_reload_from_master:
            # Just remove everything and re-add all of the items.
            # TODO make this smarter somehow to avoid flicker. Maybe change
            # this so that it removes one by one and adds back one by one.
            self.list_store_top.remove_all()
            self.list_store_bottom.remove_all()

            for e in self.list_store[:entries_before_fold]:
                self.list_store_top.append(e)

            for e in self.list_store[entries_before_fold:]:
                self.list_store_bottom.append(e)
        else:
            top_diff = len(self.list_store_top) - entries_before_fold

            if top_diff < 0:
                # Move entries from the bottom store.
                for e in self.list_store_bottom[:-top_diff]:
                    self.list_store_top.append(e)
                for _ in range(-top_diff):
                    if len(self.list_store_bottom) == 0:
                        break
                    del self.list_store_bottom[0]
            else:
                # Move entries to the bottom store.
                for e in reversed(self.list_store_top[entries_before_fold:]):
                    self.list_store_bottom.insert(0, e)
                for _ in range(top_diff):
                    del self.list_store_top[-1]

        if self.current_selection is not None:
            if not selection_changed:
                return

            self.grid_top.select_child(
                self.grid_top.get_child_at_index(self.current_selection))

            for c in self.detail_box_inner.get_children():
                self.detail_box_inner.remove(c)

            model = self.list_store[self.current_selection]
            detail_element = AlbumWithSongs(model.album, cover_art_size=300)
            detail_element.connect(
                'song-clicked',
                lambda _, song, queue, metadata: self.emit(
                    'song-clicked', song, queue, metadata),
            )
            detail_element.connect('song-selected', lambda *a: None)

            self.detail_box_inner.pack_start(detail_element, True, True, 0)
            self.detail_box.show_all()

            # TODO scroll so that the grid_top is visible, and the detail_box
            # is visible, with preference to the grid_top. May need to add
            # another flag for this function.
        else:
            self.grid_top.unselect_all()
            self.detail_box.hide()



import os
import glob
import threading
import shutil
import json
import hashlib
from collections import defaultdict
from time import sleep

from concurrent.futures import ThreadPoolExecutor, Future
from enum import EnumMeta, Enum
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    List,
    Optional,
    Union,
    Callable,
    Set,
    DefaultDict,
    Tuple,
)

import requests

from .config import AppConfiguration, ServerConfiguration
from .server import Server
from .server.api_object import APIObject
from .server.api_objects import (
    Playlist,
    PlaylistWithSongs,
    Child,
    Genre,

    # Non-ID3 versions
    Artist,
    ArtistInfo,
    Directory,

    # ID3 versions
    ArtistID3,
    ArtistInfo2,
    ArtistWithAlbumsID3,
    AlbumWithSongsID3,
)


class Singleton(type):
    def __getattr__(cls, name):
        if not CacheManager._instance:
            return None
        # If the cache has a function to do the thing we want, use it. If
        # not, then go directly to the server (this is useful for things
        # that just send data  to the server.)
        if hasattr(CacheManager._instance, name):
            return getattr(CacheManager._instance, name)
        else:
            return getattr(CacheManager._instance.server, name)

        return None


class SongCacheStatus(Enum):
    NOT_CACHED = 0
    CACHED = 1
    PERMANENTLY_CACHED = 2
    DOWNLOADING = 3


class CacheManager(metaclass=Singleton):
    executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=50)
    should_exit: bool = False

    @staticmethod
    def ready():
        return CacheManager._instance is not None

    @staticmethod
    def shutdown():
        CacheManager.should_exit = True
        print('Shutdown start')
        CacheManager.executor.shutdown()
        print('Shutdown complete')

    @staticmethod
    def calculate_server_hash(server: ServerConfiguration):
        if server is None:
            return None
        server_info = (server.name + server.server_address + server.username)
        return hashlib.md5(server_info.encode('utf-8')).hexdigest()[:8]

    class CacheEncoder(json.JSONEncoder):
        def default(self, obj):
            if type(obj) == datetime:
                return int(obj.timestamp() * 1000)
            elif type(obj) == set:
                return list(obj)
            elif isinstance(obj, APIObject):
                return {k: v for k, v in obj.__dict__.items() if v is not None}
            elif isinstance(obj, EnumMeta):
                return None

            return json.JSONEncoder.default(self, obj)

    class __CacheManagerInternal:
        # Thread lock for preventing threads from overriding the state while
        # it's being saved.
        cache_lock = threading.Lock()

        cache: DefaultDict[str, Any] = defaultdict(dict)
        permanently_cached_paths: Set[str] = set()

        # The server instance.
        server: Server
        browse_by_tags: bool

        # TODO need to split out the song downloads and make them higher
        # priority I think. Maybe even need to just make this a priority queue.
        download_set_lock = threading.Lock()
        current_downloads: Set[Path] = set()

        def __init__(
                self,
                app_config: AppConfiguration,
                server_config: ServerConfiguration,
        ):
            self.app_config = app_config
            self.browse_by_tags = self.app_config.server.browse_by_tags
            self.server_config = server_config
            self.server = Server(
                name=server_config.name,
                hostname=server_config.server_address,
                username=server_config.username,
                password=server_config.password,
                disable_cert_verify=server_config.disable_cert_verify,
            )
            self.download_limiter_semaphore = threading.Semaphore(
                self.app_config.concurrent_download_limit)

            self.load_cache_info()

        def load_cache_info(self):
            cache_meta_file = self.calculate_abs_path('.cache_meta')

            meta_json = {}
            if cache_meta_file.exists():
                with open(cache_meta_file, 'r') as f:
                    try:
                        meta_json = json.load(f)
                    except json.decoder.JSONDecodeError:
                        return

            cache_configs = [
                # Playlists
                ('playlists', Playlist, list),
                ('playlist_details', PlaylistWithSongs, dict),
                ('song_details', Child, dict),
                ('genres', Genre, list),

                # Non-ID3 caches
                ('albums', Child, 'dict-list'),
                ('album_details', Directory, dict),
                ('artists', Artist, list),
                ('artist_details', Directory, dict),
                ('artist_infos', ArtistInfo, dict),

                # ID3 caches
                ('albums_id3', AlbumWithSongsID3, 'dict-list'),
                ('album_details_id3', AlbumWithSongsID3, dict),
                ('artists_id3', ArtistID3, list),
                ('artist_details_id3', ArtistWithAlbumsID3, dict),
                ('artist_infos_id3', ArtistInfo2, dict),
            ]
            for name, type_name, default in cache_configs:
                if default == list:
                    self.cache[name] = [
                        type_name.from_json(x)
                        for x in meta_json.get(name, [])
                    ]
                elif default == dict:
                    self.cache[name] = {
                        id: type_name.from_json(x)
                        for id, x in meta_json.get(name, {}).items()
                    }
                elif default == 'dict-list':
                    self.cache[name] = {
                        n: [type_name.from_json(x) for x in xs]
                        for n, xs in meta_json.get(name, {}).items()
                    }

        def save_cache_info(self):
            os.makedirs(self.app_config.cache_location, exist_ok=True)

            cache_meta_file = self.calculate_abs_path('.cache_meta')
            os.makedirs(os.path.dirname(cache_meta_file), exist_ok=True)
            with open(cache_meta_file, 'w+') as f, self.cache_lock:
                f.write(
                    json.dumps(
                        self.cache, indent=2, cls=CacheManager.CacheEncoder))

        def save_file(self, absolute_path: Path, data: bytes):
            # Make the necessary directories and write to file.
            os.makedirs(absolute_path.parent, exist_ok=True)
            with open(absolute_path, 'wb+') as f:
                f.write(data)

        def calculate_abs_path(self, *relative_paths):
            return Path(self.app_config.cache_location).joinpath(
                CacheManager.calculate_server_hash(self.server_config),
                *relative_paths,
            )

        def calculate_download_path(self, *relative_paths):
            """
            Determine where to temporarily put the file as it is downloading.
            """
            xdg_cache_home = (
                os.environ.get('XDG_CACHE_HOME')
                or os.path.expanduser('~/.cache'))
            return Path(xdg_cache_home).joinpath(
                'sublime-music',
                CacheManager.calculate_server_hash(self.server_config),
                *relative_paths,
            )

        def return_cached_or_download(
                self,
                relative_path: Union[Path, str],
                download_fn: Callable[[], bytes],
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
                allow_download: bool = True,
        ):
            abs_path = self.calculate_abs_path(relative_path)
            download_path = self.calculate_download_path(relative_path)
            if not abs_path.exists() or force:
                if not allow_download:
                    return None

                before_download()
                resource_downloading = False
                with self.download_set_lock:
                    if abs_path in self.current_downloads:
                        resource_downloading = True

                    self.current_downloads.add(abs_path)

                if resource_downloading:
                    print(abs_path, 'already being downloaded.')
                    # The resource is already being downloaded. Busy loop until
                    # it has completed. Then, just return the path to the
                    # resource.
                    while abs_path in self.current_downloads:
                        sleep(0.5)

                    return str(abs_path)

                else:
                    print(abs_path, 'not found. Downloading...')

                    os.makedirs(download_path.parent, exist_ok=True)
                    self.save_file(download_path, download_fn())

                    # Move the file to its cache download location.
                    os.makedirs(abs_path.parent, exist_ok=True)
                    if download_path.exists():
                        shutil.move(download_path, abs_path)

                    with self.download_set_lock:
                        self.current_downloads.discard(abs_path)

                print(abs_path, 'downloaded. Returning.')

            return str(abs_path)

        @staticmethod
        def create_future(fn, *args):
            """
            Creates a future on the CacheManager's executor.
            """
            return CacheManager.executor.submit(fn, *args)

        def delete_cached_cover_art(self, id: int):
            relative_path = f'cover_art/*{id}_*'

            abs_path = self.calculate_abs_path(relative_path)

            for path in glob.glob(str(abs_path)):
                Path(path).unlink()

        def get_playlists(
                self,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_playlists() -> List[Playlist]:
                if not self.cache.get('playlists') or force:
                    before_download()
                    playlists = self.server.get_playlists().playlist
                    with self.cache_lock:
                        self.cache['playlists'] = playlists
                    self.save_cache_info()
                return self.cache['playlists']

            return CacheManager.create_future(do_get_playlists)

        def invalidate_playlists_cache(self):
            if not self.cache.get('playlists'):
                return

            with self.cache_lock:
                self.cache['playlists'] = []
            self.save_cache_info()

        def get_playlist(
                self,
                playlist_id: int,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_playlist() -> PlaylistWithSongs:
                cache_name = "playlist_details"

                if playlist_id not in self.cache.get(cache_name, {}) or force:
                    before_download()
                    playlist = self.server.get_playlist(playlist_id)
                    with self.cache_lock:
                        self.cache['playlist_details'][playlist_id] = playlist

                        # Playlists have the song details, so save those too.
                        for song in (playlist.entry or []):
                            self.cache['song_details'][song.id] = song

                    self.save_cache_info()

                playlist_details = self.cache['playlist_details'][playlist_id]

                # Invalidate the cached photo if we are forcing a retrieval
                # from the server.
                if force:
                    self.delete_cached_cover_art(playlist.id)

                return playlist_details

            return CacheManager.create_future(do_get_playlist)

        def create_playlist(self, name: str) -> Future:
            def do_create_playlist():
                self.server.create_playlist(name=name)

            return CacheManager.create_future(do_create_playlist)

        def update_playlist(self, playlist_id, *args, **kwargs):
            def do_update_playlist():
                self.server.update_playlist(playlist_id, *args, **kwargs)
                with self.cache_lock:
                    del self.cache['playlist_details'][playlist_id]

            return CacheManager.create_future(do_update_playlist)

        def get_artists(
                self,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_artists() -> List[Union[Artist, ArtistID3]]:
                cache_name = self.id3ify('artists')
                server_fn = (
                    self.server.get_artists
                    if self.browse_by_tags else self.server.get_indexes)

                if not self.cache.get(cache_name) or force:
                    before_download()
                    raw_artists = server_fn()

                    artists: List[Union[Artist, ArtistID3]] = []
                    for index in raw_artists.index:
                        artists.extend(index.artist)

                    with self.cache_lock:
                        self.cache[cache_name] = artists

                    self.save_cache_info()

                return self.cache[cache_name]

            return CacheManager.create_future(do_get_artists)

        def get_artist(
                self,
                artist_id,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_artist() -> Union[ArtistWithAlbumsID3, Child]:
                cache_name = self.id3ify('artist_details')
                server_fn = (
                    self.server.get_artist if self.browse_by_tags else
                    self.server.get_music_directory)

                if artist_id not in self.cache.get(cache_name, {}) or force:
                    before_download()
                    artist = server_fn(artist_id)

                    with self.cache_lock:
                        self.cache[cache_name][artist_id] = artist

                    self.save_cache_info()

                return self.cache[cache_name][artist_id]

            return CacheManager.create_future(do_get_artist)

        def get_artist_info(
                self,
                artist_id,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_artist_info() -> Union[ArtistInfo, ArtistInfo2]:
                cache_name = self.id3ify('artist_infos')
                server_fn = (
                    self.server.get_artist_info2
                    if self.browse_by_tags else self.server.get_artist_info)

                if artist_id not in self.cache.get(cache_name, {}) or force:
                    before_download()
                    artist_info = server_fn(id=artist_id)

                    if artist_info:
                        with self.cache_lock:
                            self.cache[cache_name][artist_id] = artist_info

                        self.save_cache_info()

                return self.cache[cache_name][artist_id]

            return CacheManager.create_future(do_get_artist_info)

        def get_artist_artwork(
                self,
                artist: Union[Artist, ArtistID3],
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_artist_artwork_filename() -> str:
                artist_info = CacheManager.get_artist_info(artist.id).result()
                lastfm_url = ''.join(artist_info.largeImageUrl)
                placeholder_image_url = 'https://lastfm-img2.akamaized.net' \
                    '/i/u/300x300/2a96cbd8b46e442fc41c2b86b821562f.png'

                # If it is the placeholder LastFM URL, then just use the cover
                # art filename given by the server.
                if lastfm_url == placeholder_image_url:
                    if isinstance(artist, ArtistWithAlbumsID3):
                        return CacheManager.get_cover_art_filename(
                            artist.coverArt, size=300).result()
                    elif (isinstance(artist, Directory)
                          and len(artist.child) > 0):
                        # Retrieve the first album's cover art
                        return CacheManager.get_cover_art_filename(
                            artist.child[0].coverArt, size=300).result()

                url_hash = hashlib.md5(lastfm_url.encode('utf-8')).hexdigest()
                return self.return_cached_or_download(
                    f'cover_art/artist.{url_hash}',
                    lambda: requests.get(lastfm_url).content,
                    before_download=before_download,
                    force=force,
                )

            return CacheManager.create_future(do_get_artist_artwork_filename)

        def get_album_list(
                self,
                type_: str,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
                # Look at documentation for get_album_list in server.py:
                **params,
        ) -> Future:
            cache_name = self.id3ify('albums')
            server_fn = (
                self.server.get_album_list2
                if self.browse_by_tags else self.server.get_album_list)

            def get_page(offset, page_size=500):
                return server_fn(
                    type_,
                    size=page_size,
                    offset=offset,
                    **params,
                ).album or []

            def do_get_album_list() -> List[Union[Child, AlbumWithSongsID3]]:
                if not self.cache.get(cache_name, {}).get(type_, []) or force:
                    before_download()

                    page_size = 40 if type_ == 'random' else 500
                    offset = 0

                    next_page = get_page(offset, page_size=page_size)
                    albums = next_page

                    # If it returns 500 things, then there's more leftover.
                    while len(next_page) == 500:
                        next_page = get_page(offset)
                        albums.extend(next_page)

                    # Update the cache.
                    with self.cache_lock:
                        if not self.cache[cache_name].get(type_):
                            self.cache[cache_name][type_] = []
                        self.cache[cache_name][type_] = albums
                    self.save_cache_info()

                return self.cache[cache_name][type_]

            return CacheManager.create_future(do_get_album_list)

        def invalidate_album_list(self, type_):
            # TODO make this invalidate instead of delete
            cache_name = self.id3ify('albums')
            if not self.cache.get(cache_name, {}).get(type_):
                return
            with self.cache_lock:
                self.cache[cache_name][type_] = []
            self.save_cache_info()

        def get_album(
                self,
                album_id,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_album() -> Union[AlbumWithSongsID3, Child]:
                cache_name = self.id3ify('album_details')
                server_fn = (
                    self.server.get_album if self.browse_by_tags else
                    self.server.get_music_directory)

                if album_id not in self.cache.get(cache_name, {}) or force:
                    before_download()
                    album = server_fn(album_id)

                    with self.cache_lock:
                        self.cache[cache_name][album_id] = album

                    self.save_cache_info()

                return self.cache[cache_name][album_id]

            return CacheManager.create_future(do_get_album)

        def batch_delete_cached_songs(
                self,
                song_ids: List[int],
                on_song_delete: Callable[[], None],
        ) -> Future:
            def do_delete_cached_songs():
                # Do the actual download.
                for song_id in song_ids:
                    song_details_future = CacheManager.get_song_details(
                        song_id)

                    def filename_future_done(f):
                        relative_path = f.result().path
                        abs_path = self.calculate_abs_path(relative_path)
                        if abs_path.exists():
                            abs_path.unlink()
                        on_song_delete()

                    song_details_future.add_done_callback(filename_future_done)

            return CacheManager.create_future(do_delete_cached_songs)

        def batch_download_songs(
                self,
                song_ids: List[int],
                before_download: Callable[[], None],
                on_song_download_complete: Callable[[int], None],
        ) -> Future:
            def do_download_song(song_id):
                # Prevents further songs from being downloaded.
                if CacheManager.should_exit:
                    return

                # Do the actual download.
                song_details_future = CacheManager.get_song_details(song_id)
                song = song_details_future.result()
                self.return_cached_or_download(
                    song.path,
                    lambda: self.server.download(song.id),
                    before_download=before_download,
                )

                self.download_limiter_semaphore.release()
                on_song_download_complete(song_id)

            def do_batch_download_songs():
                self.current_downloads = self.current_downloads.union(
                    set(song_ids))
                for song_id in song_ids:
                    self.download_limiter_semaphore.acquire()
                    CacheManager.create_future(do_download_song, song_id)

            return CacheManager.create_future(do_batch_download_songs)

        def get_cover_art_filename(
                self,
                id: str,
                before_download: Callable[[], None] = lambda: None,
                size: Union[str, int] = 200,
                force: bool = False,
                allow_download: bool = True,
        ) -> Future:
            def do_get_cover_art_filename() -> str:
                tag = 'tag_' if self.browse_by_tags else ''
                return self.return_cached_or_download(
                    f'cover_art/{tag}{id}_{size}',
                    lambda: self.server.get_cover_art(id, str(size)),
                    before_download=before_download,
                    force=force,
                    allow_download=allow_download,
                )

            return CacheManager.create_future(do_get_cover_art_filename)

        def get_song_details(
                self,
                song_id: int,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_song_details() -> Child:
                if not self.cache['song_details'].get(song_id) or force:
                    before_download()
                    song_details = self.server.get_song(song_id)
                    with self.cache_lock:
                        self.cache['song_details'][song_id] = song_details
                    self.save_cache_info()

                return self.cache['song_details'][song_id]

            return CacheManager.create_future(do_get_song_details)

        def get_play_queue(self) -> Future:
            return CacheManager.create_future(self.server.get_play_queue)

        def save_play_queue(
                self,
                play_queue: List[str],
                current: str,
                position: float,
        ):
            CacheManager.create_future(
                self.server.save_play_queue, play_queue, current, position)

        def scrobble(self, song_id: int) -> Future:
            def do_scrobble():
                self.server.scrobble(song_id)

            return CacheManager.create_future(do_scrobble)

        def get_song_filename_or_stream(
                self,
                song: Child,
                format=None,
                force_stream: bool = False,
        ) -> Tuple[str, bool]:
            abs_path = self.calculate_abs_path(song.path)
            if not abs_path.exists() or force_stream:
                return (
                    self.server.get_stream_url(song.id, format=format),
                    True,
                )

            return (str(abs_path), False)

        def get_genres(
                self,
                before_download: Callable[[], None] = lambda: None,
                force: bool = False,
        ) -> Future:
            def do_get_genres() -> List[Genre]:
                if not self.cache['genres'] or force:
                    before_download()
                    genres = self.server.get_genres().genre
                    with self.cache_lock:
                        self.cache['genres'] = genres
                    self.save_cache_info()

                return self.cache['genres']

            return CacheManager.create_future(do_get_genres)

        def get_cached_status(self, song: Child) -> SongCacheStatus:
            cache_path = self.calculate_abs_path(song.path)
            if cache_path.exists():
                if cache_path in self.permanently_cached_paths:
                    return SongCacheStatus.PERMANENTLY_CACHED
                else:
                    return SongCacheStatus.CACHED
            elif cache_path in self.current_downloads:
                return SongCacheStatus.DOWNLOADING
            else:
                return SongCacheStatus.NOT_CACHED

        def id3ify(self, cache_type):
            return cache_type + ('_id3' if self.browse_by_tags else '')

    _instance: Optional[__CacheManagerInternal] = None

    def __init__(self, server_config: ServerConfiguration):
        raise Exception('Do not instantiate the CacheManager.')

    @classmethod
    def reset(cls, app_config, server_config):
        CacheManager._instance = CacheManager.__CacheManagerInternal(
            app_config, server_config)
